@@ -3,11 +3,9 @@ from pathlib import Path
 
 cwd = os.getcwd()
 tmp_dir = cwd + "/tmp"
-data_dir = tmp_dir + "/data"
 logger_dir = cwd + "/log"
 
 os.makedirs(tmp_dir, exist_ok=True)
-os.makedirs(data_dir, exist_ok=True)
 os.makedirs(logger_dir, exist_ok=True)
 
 
