@@ -61,7 +61,7 @@ from lecrapaud.integrations.openai_integration import (
 )
 from lecrapaud.feature_selection import get_features_by_types
 from lecrapaud.utils import logger
-from lecrapaud.db import Target, Feature, Dataset
+from lecrapaud.db import Target, Feature, Experiment
 from lecrapaud.config import PYTHON_ENV
 
 
@@ -308,7 +308,7 @@ class PreprocessFeature:
     def __init__(
         self,
         data: pd.DataFrame,
-        dataset,
+        experiment,
         time_series: bool = False,
         date_column: str | None = None,
         group_column: str | None = None,
@@ -326,7 +326,7 @@ class PreprocessFeature:
         self.data = data
         self.data.columns = self.data.columns.str.upper()
 
-        self.dataset = dataset
+        self.experiment = experiment
         self.columns_pca = columns_pca
         self.columns_onehot = columns_onehot
         self.columns_binary = columns_binary
@@ -341,10 +341,10 @@ class PreprocessFeature:
         self.val_size = val_size
         self.test_size = test_size
 
-        self.dataset_dir = self.dataset.path
-        self.dataset_id = self.dataset.id
-        self.data_dir = f"{self.dataset_dir}/data"
-        self.preprocessing_dir = f"{self.dataset_dir}/preprocessing"
+        self.experiment_dir = self.experiment.path
+        self.experiment_id = self.experiment.id
+        self.data_dir = f"{self.experiment_dir}/data"
+        self.preprocessing_dir = f"{self.experiment_dir}/preprocessing"
 
     def run(self):
         # Split
@@ -358,8 +358,8 @@ class PreprocessFeature:
 
         # PCA
         train, pcas = self.add_pca_features(train)
-        val, _ = self.add_pca_features(test, pcas=pcas)
-        test, _ = self.add_pca_features(val, pcas=pcas)
+        val, _ = self.add_pca_features(val, pcas=pcas)
+        test, _ = self.add_pca_features(test, pcas=pcas)
 
         joblib.dump(pcas, f"{self.preprocessing_dir}/pcas.pkl")
 
@@ -377,7 +377,7 @@ class PreprocessFeature:
         joblib.dump(self.data, f"{self.data_dir}/full.pkl")
         joblib.dump(transformer, f"{self.preprocessing_dir}/column_transformer.pkl")
         summary = summarize_dataframe(train)
-        summary.to_csv(f"{self.dataset_dir}/feature_summary.csv", index=False)
+        summary.to_csv(f"{self.experiment_dir}/feature_summary.csv", index=False)
 
         return train, val, test
 
@@ -431,9 +431,9 @@ class PreprocessFeature:
                 f"{data.shape} {name} data from {dates[f"{name}_start_date"].strftime('%d/%m/%Y')} to {dates[f"{name}_end_date"].strftime('%d/%m/%Y')}"
             )
 
-        Dataset.update(
+        Experiment.upsert(
             match_fields=["id"],
-            id=self.dataset_id,
+            id=self.experiment_id,
             train_size=len(train),
             val_size=len(val),
             test_size=len(test),
@@ -454,7 +454,7 @@ class PreprocessFeature:
         Splits a DataFrame into train, validation, and test sets.
 
         Parameters:
-            df (pd.DataFrame): The full dataset
+            df (pd.DataFrame): The full experiment
             val_size (float): Proportion of validation set (default 0.1)
             test_size (float): Proportion of test set (default 0.1)
             random_state (int): Random seed for reproducibility
@@ -646,23 +646,52 @@ class PreprocessFeature:
 
         # Insert features in db
         if save_in_db:
-            # TODO: in bulk
+            # Get feature types from transformed data
             categorical_features, numerical_features = get_features_by_types(
                 X_transformed
             )
-            for feature in categorical_features:
-                Feature.upsert(match_fields=["name"], name=feature, type="categorical")
-            for feature in numerical_features:
-                Feature.upsert(match_fields=["name"], name=feature, type="numerical")
-            for target in y.columns:
-                target_number = int(target.split("_")[1])
-                type = (
+
+            # Get column names from DataFrames
+            cat_feature_names = categorical_features.columns.tolist()
+            num_feature_names = numerical_features.columns.tolist()
+
+            # Combine all feature names and their types
+            all_feature_names = cat_feature_names + num_feature_names
+            all_feature_types = ["categorical"] * len(cat_feature_names) + [
+                "numerical"
+            ] * len(num_feature_names)
+
+            # Upsert features in bulk if we have any features
+            if all_feature_names:
+                Feature.upsert_bulk(
+                    match_fields=["name"],
+                    name=all_feature_names,
+                    type=all_feature_types,
+                )
+
+            # Upsert targets in bulk
+            target_names = y.columns.tolist()
+            target_types = [
+                (
                     "classification"
-                    if target_number in self.target_clf
+                    if int(target.split("_")[1]) in self.target_clf
                     else "regression"
                 )
-                # TODO: what about description here ?
-                Target.upsert(match_fields=["name", "type"], name=target, type=type)
+                for target in target_names
+            ]
+
+            Target.upsert_bulk(
+                match_fields=["name"], name=target_names, type=target_types
+            )
+
+            # Get all the upserted objects
+            targets = Target.filter(name__in=target_names)
+
+            # Update experiment with targets
+            experiment = Experiment.get(self.experiment_id)
+            if experiment:
+                experiment.targets = targets
+                experiment.save()
 
         return pd.concat([X_transformed, y], axis=1), transformer
 
