@@ -41,7 +41,7 @@ from lecrapaud.directories import tmp_dir, clean_directory
 from lecrapaud.utils import logger
 from lecrapaud.config import PYTHON_ENV
 from lecrapaud.db import (
-    Dataset,
+    Experiment,
     Target,
     Feature,
     FeatureSelection,
@@ -54,8 +54,8 @@ from lecrapaud.search_space import all_models
 warnings.filterwarnings("ignore", category=FutureWarning)
 
 
-def load_train_data(dataset_dir, target_number, target_type="regression"):
-    data_dir = f"{dataset_dir}/data"
+def load_train_data(experiment_dir, target_number, target_type="regression"):
+    data_dir = f"{experiment_dir}/data"
 
     logger.info("Loading data...")
     train = joblib.load(f"{data_dir}/train.pkl")
@@ -74,8 +74,8 @@ def load_train_data(dataset_dir, target_number, target_type="regression"):
 
 
 class FeatureSelectionEngine:
-    def __init__(self, train, dataset, target_number, target_clf, **kwargs):
-        self.dataset = dataset
+    def __init__(self, train, experiment, target_number, target_clf, **kwargs):
+        self.experiment = experiment
         self.train = train
         self.target_number = target_number
         self.target_clf = target_clf
@@ -83,16 +83,16 @@ class FeatureSelectionEngine:
         self.target_type = (
             "classification" if self.target_number in self.target_clf else "regression"
         )
-        self.percentile = self.dataset.percentile
-        self.corr_threshold = self.dataset.corr_threshold
-        self.max_features = self.dataset.max_features
+        self.percentile = self.experiment.percentile
+        self.corr_threshold = self.experiment.corr_threshold
+        self.max_features = self.experiment.max_features
 
-        self.dataset_dir = self.dataset.path
-        self.dataset_id = self.dataset.id
-        self.data_dir = f"{self.dataset_dir}/data"
-        self.preprocessing_dir = f"{self.dataset_dir}/preprocessing"
+        self.experiment_dir = self.experiment.path
+        self.experiment_id = self.experiment.id
+        self.data_dir = f"{self.experiment_dir}/data"
+        self.preprocessing_dir = f"{self.experiment_dir}/preprocessing"
         self.fs_dir_target = (
-            f"{self.dataset_dir}/{f"TARGET_{self.target_number}"}/feature_selection"
+            f"{self.experiment_dir}/{f"TARGET_{self.target_number}"}/feature_selection"
         )
         os.makedirs(self.fs_dir_target, exist_ok=True)
 
@@ -119,12 +119,14 @@ class FeatureSelectionEngine:
         max_features = self.max_features
 
         feature_selection = FeatureSelection.upsert(
-            match_fields=["target_id", "dataset_id"],
+            match_fields=["target_id", "experiment_id"],
             target_id=target.id,
-            dataset_id=self.dataset_id,
+            experiment_id=self.experiment_id,
         )
 
-        if feature_selection.best_features_path:
+        if feature_selection.best_features_path and os.path.exists(
+            feature_selection.best_features_path
+        ):
             return joblib.load(feature_selection.best_features_path)
 
         self.X = self.train.loc[:, ~self.train.columns.str.contains("^TARGET_")]
@@ -767,7 +769,7 @@ class PreprocessModel:
         train,
         val,
         test,
-        dataset,
+        experiment,
         target_numbers,
         target_clf,
         models_idx,
@@ -777,7 +779,10 @@ class PreprocessModel:
         date_column,
         **kwargs,
     ):
-        self.dataset = dataset
+        self.train = train
+        self.val = val
+        self.test = test
+        self.experiment = experiment
         self.target_numbers = target_numbers
         self.target_clf = target_clf
         self.models_idx = models_idx
@@ -786,13 +791,16 @@ class PreprocessModel:
         self.group_column = group_column
         self.date_column = date_column
 
-        self.dataset_dir = dataset.path
-        self.data_dir = f"{self.dataset_dir}/data"
-        self.preprocessing_dir = f"{self.dataset_dir}/preprocessing"
+        self.experiment_dir = experiment.path
+        self.data_dir = f"{self.experiment_dir}/data"
+        self.preprocessing_dir = f"{self.experiment_dir}/preprocessing"
 
-        self.all_features = dataset.get_all_features(
+        self.all_features = experiment.get_all_features(
             date_column=date_column, group_column=group_column
         )
+
+    def run(self):
+        # save data
         columns_to_keep = self.all_features + [
             f"TARGET_{i}" for i in self.target_numbers
         ]
@@ -801,15 +809,9 @@ class PreprocessModel:
         ]
         if duplicates:
             raise ValueError(f"Doublons détectés dans columns_to_keep: {duplicates}")
-
-        self.train = train[columns_to_keep]
-        if isinstance(val, pd.DataFrame):
-            self.val = val[columns_to_keep]
-        if isinstance(test, pd.DataFrame):
-            self.test = test[columns_to_keep]
-
-    def run(self):
-        # save data
+        self.train = self.train[columns_to_keep]
+        self.val = self.val[columns_to_keep]
+        self.test = self.test[columns_to_keep]
         joblib.dump(self.train, f"{self.data_dir}/train.pkl")
         joblib.dump(self.val, f"{self.data_dir}/val.pkl")
         joblib.dump(self.test, f"{self.data_dir}/test.pkl")
@@ -870,8 +872,11 @@ class PreprocessModel:
 
     def inference(self):
         # self.train is new data here
+        columns_to_keep = self.all_features
+        self.train = self.train[columns_to_keep]
+
         scaler_x = joblib.load(f"{self.preprocessing_dir}/scaler_x.pkl")
-        scaled_data = scaler_x.transform(self.train)
+        scaled_data = scaler_x.transform(self.train)  # TODO: utiliser scale_data
         scaled_data = pd.DataFrame(
             scaled_data, columns=self.train.columns, index=self.train.index
         )
@@ -1012,7 +1017,7 @@ class PreprocessModel:
             df_reshaped = df.apply(list, axis=1).apply(lambda x: [list(x)])
             df_reshaped = pd.concat([df_reshaped, group_series], axis=1)
 
-            logger.info("Grouping method stock and creating timesteps...")
+            logger.info("Grouping features and creating timesteps...")
             df_reshaped = (
                 df_reshaped.groupby(group_column)[0]
                 .apply(lambda x: shiftsum(x, timesteps))
@@ -1058,13 +1063,13 @@ class PreprocessModel:
 def feature_selection_analysis(feature_selection_id: int, n_components: int = 5):
 
     feature_selection = FeatureSelection.get(feature_selection_id)
-    dataset_dir = feature_selection.dataset.path
+    experiment_dir = feature_selection.experiment.path
     features = [f.name for f in feature_selection.features]
     target = feature_selection.target.name
     target_number = target.split("_")[1]
 
     train, val, train_scaled, val_scaled, _scaler_y = load_train_data(
-        dataset_dir, target_number, target_type=feature_selection.target.type
+        experiment_dir, target_number, target_type=feature_selection.target.type
     )
     train = train[features + [target]]
     train_scaled = train_scaled[features + [target]]
@@ -1135,6 +1140,7 @@ def get_features_by_types(df: pd.DataFrame, sample_categorical_threshold: int =
         for col in df.columns
         if df[col].nunique() <= sample_categorical_threshold
         and df[col].dtype in ["int64", "Int64"]
+        and (df[col] >= 0).all()
     ]
     df_categorical = df[categorical_features]
     logger.info(f"Number of categorical features: {len(categorical_features)}")
